{% extends "!layout.html" %}
{% block sidebartitle %}
{{ super() }}
<a href="https://github.com/pyca/cryptography" style="display:block;margin-top:8px;text-align:center;">View on GitHub</a>
{% endblock %}
